#! /usr/bin/env python3

import pysat
import collections, bz2
import multiprocessing
import datetime, time
#max_years = 1  # Ref: https://github.com/python/cpython/blob/889b0b9bf95651fc05ad532cc4a66c0f8ff29fc2/Include/cpython/pytime.h
#max_sleep_time = datetime.timedelta(days=1).total_seconds()
import sys
sys.path.append('/auto/msioutis/qcn2sat/lib/python3.5/site-packages/')

class QCN:
    """Qualitative Constraint Network (restricted to arcs i < j)"""
    def __init__(self, signature):
        self.size = 0
        self.signature = frozenset(signature)
        s_sign = list(signature)
        s_sign.sort()
        self.universal = s_sign
        self.constraints = dict()
        self.graph = None

    def add(self, i, j, relation):
        """"Add a constraint (i < j)"""
        assert i < j
        for base_rel in relation:
            if not base_rel in self.signature:
                raise SystemExit("Base relation not in given signature\n")

        idx = (i, j)
        old = self.signature
        if idx in self.constraints:
            old = frozenset(self.constraints[idx])
        relation_sort = list(frozenset(relation) & old)
        relation_sort.sort()

        self.constraints[idx] = relation_sort

    def get(self, i, j):
        """Read constraint on (i,j); always returns a relation."""
        try:
            return self.constraints[(i, j)]
        except KeyError:
            return self.universal

    def iterate_strict_triangle(self):
        """Iterate all pairs i < j on current primal constraint graph."""
        iterate = [ (i, j) for (i, j) in self.graph ]
        iterate.sort()
        for (i, j) in iterate:
            yield i, j


    def iterate_strict_triples(self):
        """iterate i < j < k triples"""
        iterate = [ (i, j) for (i, j) in self.graph ]
        iterate.sort()

        neighb = dict()
        for i in xrange(self.size):
            neighb[i] = set()

        for (i, j) in iterate:
            neighb[i].add(j)

        for (i, j) in iterate:
            for k in neighb[i] & neighb[j]:
                yield i, j, k

class PropositionalAtoms: 
    """Dictionary enumerating variable strings."""
    def __init__(self):
        self.names = dict()
        self.aux = dict()
        self.last_used = 0
    
    def encode(self, i, j, baserel):
        try:
            return self.names[i][j][baserel]
        except KeyError:
            assert i <= j
            self.last_used += 1

            if not i in self.names:
                self.names[i] = dict()
            if not j in self.names[i]:
                self.names[i][j] = dict()
            self.names[i][j][baserel] = self.last_used
            return self.last_used
            
    def encode_aux(self, i, j):
        try:
            return self.aux[i][j]
        except KeyError:
            assert i <= j
            self.last_used += 1

            if not i in self.aux:
                self.aux[i] = dict()
            self.aux[i][j] = self.last_used
            return self.last_used

def read_comp_table(calculus):
    """Read an GQR-format composition table"""
    import re

    table = { }
    all_relations = set()

    lines = open('data/' + calculus + '.comp', 'r')
    for line in lines:
        extract = re.search('^(.*):(.*)::[ ]*\((.*)\)$', line)
        assert(extract)
        left = extract.group(1).strip()
        right = extract.group(2).strip()
        supp = extract.group(3).strip()
        table[left + " " + right] = frozenset(supp.split())
        all_relations.add(left)
        all_relations.add(right)
    lines.close()

    return table, frozenset(all_relations)

def read_gqr_csp(buffer, signature):
    """Read an GQR-format CSP from input"""
    import re

    data = QCN(signature)
     
    info = buffer.pop(0).split()
    data.size = int(info[0].strip()) + 1
    ID = info[1].strip()

    for line in buffer:
        res = re.search('^[ ]*([0-9]*)[ ]*([0-9]*) \((.*)\)', line)
        if res:
            i = int(res.group(1))
            j = int(res.group(2))
            if i == j: continue
            assert 0 <= i < j
            rel = res.group(3).strip().split(" ")
            if not rel:
                return None
            for base_rel in rel:
                if not base_rel in signature:
                    raise SystemExit("Calculus signature does not match CSP")

            if frozenset(rel) != signature:  # ignore universal statements
                data.add(i, j, rel)

    return data, ID


def instantiate_rcc8(l, x, y, atoms): # pylint: disable=C0103
    """encode instantiated literal l on x,y"""

    assert l[1].relation in [ 'C', 'O', 'NTP', 'P' ]

    mod = 1
    if not l[0]:  # negated predicate
        mod = -1

    # symmetry
    if l[1].relation in [ 'O', 'C' ]:
        if y < x:
            rel_string = l[1].swap_variables_string()
            rel_string = rel_string.replace('x', str(x))
            rel_string = rel_string.replace('y', str(y))
            return mod * atoms.encode(y, x, rel_string)
        else:
            rel_string = l[1].string.replace('x', str(x))
            rel_string = rel_string.replace('y', str(y))
            return mod * atoms.encode(x, y, rel_string)

    # relation is 'P' or 'NTP'
    rel_string = l[1].string.replace('x', str(x))
    rel_string = rel_string.replace('y', str(y))

    if x < y:
        return mod * atoms.encode(x, y, rel_string)
    # PropositionalAtoms::encode() assumes a <= b
    # The string is NOT rewritten ... It just gets filed under (y,x)
    return mod * atoms.encode(y, x, rel_string)

def instantiate_up_to_z(l, x, y, z, atoms): # pylint: disable=C0103
    """Wrapper for instantiate supporting 3 vars"""
    from syntactic_map import Predicate

    if 'z' != l[1].var1 and 'z' != l[1].var2:
        return instantiate_rcc8(l, x, y, atoms)

    if 'x' != l[1].var1 and 'x' != l[1].var2:
        name = l[1].string
        name = name.replace('y', 'x')
        name = name.replace('z', 'y')
        return instantiate_rcc8((l[0], Predicate(name)), y, z, atoms)
    if 'y' != l[1].var1 and 'y' != l[1].var2:
        name = l[1].string
        name = name.replace('z', 'y')
        return instantiate_rcc8((l[0], Predicate(name)), x, z, atoms)
    assert False


def rcc8_rcc4_encode_input(qcn, atoms, wcnf):
    """Encode instance according to RCC8->RCC4 map (but not the RCC4-theory)."""

    import os.path
    from syntactic_map import read_map
    syntactic_interpretation = \
        read_map(os.path.join('data', 'rcc8_rcc4.map'))

    for i, j in qcn.iterate_strict_triangle():        
        relation = qcn.get(i, j)
        if sorted(relation) != qcn.universal:
           aux_var = atoms.encode_aux(i, j)

           for clause in syntactic_interpretation[frozenset(relation)]:
               new_clause = [ ]
               for l in clause:  # pylint: disable=C0103
                   new_clause.append( instantiate_rcc8(l, i, j, atoms) )
               wcnf.append( new_clause + [aux_var])
                    
           wcnf.append( [-1 * aux_var], weight=1)



def rcc8_rcc4_encode_theory(qcn, atoms, wcnf):
    """Ground the RCC8-RCC4 theory on qcn"""

    # encode RCC8 Horn theory using RCC4
    from syntactic_map import Predicate

    # 1-consistency asserted by SCRIPT

    clauses_2 = [ [ (True, Predicate("x C y")), (False, Predicate("x O y")) ],
                  [ (True, Predicate("x O y")), (False, Predicate("x P y")) ],
                  [ (True, Predicate("x P y")), (False, Predicate("x NTP y")) ],
                  [ (False, Predicate("y P x")), (False, Predicate("x NTP y")) ]
            ]

    for i, j in qcn.iterate_strict_triangle():
        assert i < j

        for rule_2 in clauses_2:
            clause = [ instantiate_rcc8(pred, i, j, atoms) for pred in rule_2 ]
            wcnf.append(clause)
            clause_sym = [ instantiate_rcc8(pred, j, i, atoms) for pred in rule_2 ]
            if frozenset(clause) != frozenset(clause_sym):
                wcnf.append(clause_sym)

    clauses_3 = [ [ (True, Predicate("x C y")),
        (False, Predicate("x C z")), (False, Predicate("z P y")) ],
        [ (True, Predicate("x NTP z")),
        (False, Predicate("x NTP y")), (False, Predicate("y P z")) ],
        [ (True, Predicate("x NTP z")),
        (False, Predicate("x P y")), (False, Predicate("y NTP z")) ],
        [ (True, Predicate("x O y")),
        (False, Predicate("z NTP y")), (False, Predicate("x C z")) ],
        [ (True, Predicate("x O y")),
        (False, Predicate("x O z")), (False, Predicate("z P y")) ],
        [ (True, Predicate("x P y")),
        (False, Predicate("x P z")), (False, Predicate("z P y")) ],
        ]

    for i in range(0, qcn.size):
        for j in range(0, qcn.size):
            if i == j or (qcn.graph and tuple(sorted((i, j))) not in qcn.graph):
                continue
            for k in range(0, qcn.size):
                if i == k or (qcn.graph and tuple(sorted((i, k))) not in qcn.graph):
                    continue
                if j == k or (qcn.graph and tuple(sorted((j, k))) not in qcn.graph):
                    continue

                for rule in clauses_3:
                    clause = [
                        instantiate_up_to_z(pred, i, j, k, atoms)
                        for pred in rule ]

                    wcnf.append(clause)
                    
def instantiate_allen(l, x, y, atoms): # pylint: disable=C0103
    """encode instantiated literal l on x,y"""

    mod = 1
    if not l[0]:  # negated predicate
        mod = -1

    if l[1].relation == '=':  # for "=" we enforce symmetry
        if y < x or (y == x and l[1].var1[1] == '+' and l[1].var2[1] == '-'):
            rel_string = l[1].swap_variables_string()
            rel_string = rel_string.replace('x', str(x))
            rel_string = rel_string.replace('y', str(y))
            return mod * atoms.encode(y, x, rel_string)

    rel_string = l[1].string.replace('x', str(x))
    rel_string = rel_string.replace('y', str(y))

    if x < y:
        return mod * atoms.encode(x, y, rel_string)
    # PropositionalAtoms::encode() assumes a <= b
    # The string is NOT rewritten ... It just gets filed under (y,x)
    return mod * atoms.encode(y, x, rel_string)


def ord_horn_encode_input(qcn, atoms, wcnf):
    """Encode instance according to ORD-Horn map (but not the ORD-theory)."""

    import os.path
    from syntactic_map import read_map
    syntactic_interpretation = \
        read_map(os.path.join('data', 'ia_ordclauses.map'))

    used_points = set()
            
    for i, j in qcn.iterate_strict_triangle():
        relation = qcn.get(i, j)
        if sorted(relation) != qcn.universal:
           aux_var = atoms.encode_aux(i, j)

           for clause in syntactic_interpretation[frozenset(relation)]:
               new_clause = []
               for l in clause:  # pylint: disable=C0103
                   new_clause.append( instantiate_allen(l, i, j, atoms) )

                   if 'x+' in l[1].string:
                       used_points.add( (i,'+') )
                   if 'x-' in l[1].string:
                       used_points.add( (i,'-') )
                   if 'y+' in l[1].string:
                       used_points.add( (j,'+') )
                   if 'y-' in l[1].string:
                       used_points.add( (j,'-') )
               wcnf.append( new_clause + [aux_var] )
               
           wcnf.append( [-1 * aux_var], weight=1 )

    return used_points

def ord_horn_encode_variables(qcn, atoms, wcnf): # pylint: disable=R0912
    """Encode instance according to ORD-Horn map."""

    # encode input
    used_points = ord_horn_encode_input(qcn, atoms, wcnf)

    # encode ORD theory
    from syntactic_map import Predicate
    for i, s1 in used_points: # pylint: disable=C0103
        # domain formula
        if s1 == '-' and (i, '+') in used_points:
            wcnf.append([ instantiate_allen((True, Predicate(None,
                'x-', '<=', 'x+')), i, i, atoms) ])
            wcnf.append([ instantiate_allen((False, Predicate(None,
                'x-', '=', 'x+')), i, i, atoms) ])
        continue

    for i, s1 in used_points: # pylint: disable=C0103
        for j, s2 in used_points: # pylint: disable=C0103
            if (i, s1) == (j, s2):
                continue

            if (not i == j) and qcn.graph and tuple(sorted((i, j))) not in qcn.graph:
                continue

            # x <= y ^ y <= x -> x = y
            if i < j or (i == j and s1 < s2):
                # avoid generating the same clauses twice
                wcnf.append([ instantiate_allen( (False, Predicate(None,
                    'x'+s1, '<=', 'y'+s2)), i, j, atoms),
                    instantiate_allen( (False, Predicate(None, 'x'+s2, '<=', 'y'+s1)),
                    j, i, atoms),
                    instantiate_allen( (True, Predicate(None, 'x'+s1, '=', 'y'+s2)),
                    i, j, atoms) ])

            # x = y -> x <= y
            wcnf.append([ instantiate_allen( (False, Predicate(None,
                'x'+s1, '=', 'y'+s2)), i, j, atoms),
                instantiate_allen( (True, Predicate(None, 'x'+s1, '<=', 'y'+s2)),
                i, j, atoms) ])
    for i, s1 in used_points:  # pylint: disable=C0103
        for j, s2 in used_points:  # pylint: disable=C0103
            if (not i == j) and qcn.graph and tuple(sorted((i, j))) not in qcn.graph:
                continue

            if (i, s1) == (j, s2):
                continue

            for k, s3 in used_points:  # pylint: disable=C0103
                if (not i == k) and qcn.graph and tuple(sorted((i, k))) not in qcn.graph:
                    continue

                if (not j == k) and qcn.graph and tuple(sorted((j, k))) not in qcn.graph:
                    continue

                if (i, s1) == (k, s3) or (j, s2) == (k, s3):
                    continue

                # x <= y ^ y <= z -> x <= z
                wcnf.append([ instantiate_allen( (False, Predicate(None,
                    'x'+s1, '<=', 'y'+s2)), i, j, atoms),
                    instantiate_allen( (False, Predicate(None, 'x'+s2, '<=', 'y'+s3)),
                    j, k, atoms),
                    instantiate_allen( (True, Predicate(None, 'x'+s1, '<=', 'y'+s3)),
                    i, k, atoms) ])              
    
def encoding(qcn, calculus, wcnf):
    
    # encode input
    
    if calculus == "allen": 
       atoms = PropositionalAtoms()
       ord_horn_encode_variables(qcn, atoms, wcnf)
    
    if calculus == "rcc8":
       atoms = PropositionalAtoms()
       rcc8_rcc4_encode_input(qcn, atoms, wcnf)
       rcc8_rcc4_encode_theory(qcn, atoms, wcnf)                 

def check_options():
    """Parse options with assertions"""
    from argparse import ArgumentParser

    parser = ArgumentParser()

    parser.add_argument("-c", "--calculus", action="store", required=False, type=str, dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %(default)s]")
    parser.add_argument("-f", "--file", action="store", required=True, type=str, dest="filename")
    parser.add_argument("-t", "--timeout", action="store", required=False, type=int, dest="timeout")

    result = parser.parse_args()

    return result.filename, result.timeout, result.calc
        
def foo(wcnf, ID):
    with RC2S(wcnf) as rc2s:
       model = rc2s.compute()
       if model:
          print("QCN:", ID, rc2s.cost, rc2s.oracle_time())
          #print("Model:", [i for i in model if i>0])
       else:
          print("QCN:", ID, "no solution")   
       rc2s.delete()           

if __name__ == '__main__':
    #start = time.process_time() # or perf_counter() 

    QCN_FILENAME, TIMEOUT, CALCULUS = check_options()

    _, SIGNATURE = read_comp_table(CALCULUS)
    
    from pysat.examples.rc2 import RC2 as RC2S
    from pysat.formula import WCNF
    from triangulation import MCS, FIC
    
    with open(QCN_FILENAME) as f:
       
       total = 0
       meter = 0
       buffer = []

       for i in f:
          buffer.append(i)

          if i.strip() == '.':
             INPUT_QCN, ID = read_gqr_csp(buffer, SIGNATURE)
             buffer.clear()

             if INPUT_QCN.size == 0:
                # no constraints read (assume problem was unsatisfiable) or
                # problem is not 2-consistency (=> unsat)
                print("p cnf 1 2")
                print("1 0")
                print("-1 0")
                raise SystemExit()

             VERTICES = set( [ i for i in range(INPUT_QCN.size ) ] )
             EDGES = tuple( [ set([]) for i in range(len(VERTICES)) ] )

             for (I, J) in INPUT_QCN.constraints.keys():
                 EDGES[I].add(J)
                 EDGES[J].add(I)

             a, a_ = MCS(VERTICES, EDGES)
             INPUT_QCN.graph = set(FIC(VERTICES, EDGES, a, a_))
             
             wcnf = WCNF()    

             encoding(INPUT_QCN, CALCULUS, wcnf)
       
             p = multiprocessing.Process(target=foo, name="Foo", args=(wcnf, ID))
             p.start()
             p.join(TIMEOUT)
             
             if p.is_alive():
                print("QCN:", ID, "timeout") 

                # Terminate foo
                p.terminate()
                p.join()
    #print("Time:", time.process_time() - start)
