#! /usr/bin/env python3

class Predicate:
    """Relation with arguments."""
    def __init__(self, string, var1=None, rel=None, var2=None):
        if string is None:
            self.var1 = var1
            self.relation = rel
            self.var2 = var2
            self.string = "%s %s %s" % (var1, rel, var2)
        else:
            import re

            match = re.match(r'^([xyz+-]+) ([<=>A-Z]+) ([xyz+-]+)$', string)

            rel = match.group(2)

            self.var1 = match.group(1)
            self.relation = rel
            self.var2 = match.group(3)
            self.string = string

        self.hashvalue = self.var1 + self.var2 + self.relation
        self.hashvalue = self.hashvalue.__hash__()

    def swap_variables_string(self):
        """Swap arguments for relation, but not the relation itself."""
        return "%s %s %s" % (self.var2, self.relation, self.var1)

    def __eq__(self, other):
        return self.var1 == other.var1 and self.var2 == other.var2 and \
            self.relation == other.relation

    def __hash__(self):
        return self.hashvalue

def read_map(filename):
    """Read map from 'filename', return dict."""
    import re

    content = open(filename, 'r')

    # map style
    regexp = re.compile(r'^x \((.*)\) y :: {(.*)}$')

    syn_map = dict()
    for line in content:
        match = re.match(regexp, line)
        if match:
            org_string = match.group(1)
            relation = frozenset(org_string.strip().split(' '))
            if relation in syn_map:
                msg = "Relation symbol '%s' appears twice in map" % (relation)
                raise SystemExit(msg)
            assert not relation in syn_map

            defining_formula = match.group(2).strip()

            cnf = parse_cnf_string(defining_formula)

            syn_map[relation] = cnf
        else:
            msg = "Failed to parse syntactic map in '%s', line '%s'" % (
                filename, line)
            raise SystemExit(msg)

    content.close()

    return syn_map

def parse_cnf_string(string):
    """Parse string denoting a CNF (defining) formula."""
    import re

    clause_regexp = re.compile(r'^{([^}]+)}(.*)')

    string = string.strip()

    clauses = [ ]
    while string:
        clause = re.match(clause_regexp, string)
        clauses.append(clause.group(1).strip())
        string = clause.group(2).strip()

    atom_regexp = re.compile(r'^([+-])\(([^)]+)\)(.*)')

    cnf = set()
    for clause in clauses:
        clause_set = set()
        while clause:
            match = re.match(atom_regexp, clause)
            if not match:
                raise SystemExit("Failed to parse clause '%s'" % (clause))
            clause = match.group(3).strip()

            positive = True
            if match.group(1) == '-':
                positive = False

            clause_set.add( (positive, Predicate(match.group(2))) )
        cnf.add( frozenset(clause_set) )

    return cnf

def is_horn_clause(clause):
    """Is clause a Horn-clause?"""

    pos = False
    for atom in clause:
        if atom[0]:
            if pos:
                return False
            pos = True
    return True

def is_horn_formula(formula):
    """Is formula a horn formula?"""

    for clause in formula:
        if not is_horn_clause(clause):
            return False
    return True

def is_primitive_formula(formula):
    """Is formula purely conjunctive?"""

    for clause in formula:
        if len(clause) > 1:
            return False
    return True
